import { writePaddedLine } from '../../utils/console'
import exec from '../../utils/exec'
import chalk from 'chalk'
import path from 'path'
import _ from 'lodash'
import fs from 'fs'

let cachedConfig = null

let cachedApps = null

const appsPath = path.join('apps')

const configPath = path.resolve('maha.js')

const _readConfig = () => {

  if(cachedConfig) return cachedConfig

  cachedConfig = require(configPath)

  return cachedConfig

}

const _writeConfig = (config) => {

  const data = `module.exports = ${JSON.stringify(config, null, 2)}`

  fs.writeFileSync(configPath, data, 'utf8')

}

const _getApps = () => {

  if(cachedApps) return cachedApps

  const config = _readConfig()

  cachedApps = config.apps

  return cachedApps

}

const _validateApp = (name) => {

  const apps = Object.keys(_getApps())

  if(!_.includes(apps, name)) throw new Error(`app ${name} is not installed`)

}

const executeForOneOrAll = async (handler, name, arg) => {

  const argument = (typeof arg === 'string') ? arg : null

  if(typeof name === 'string' && name !== 'all') {

    _validateApp(name)

    return await handler(name, argument)

  }

  const apps = _getApps()

  await Promise.mapSeries(Object.keys(apps).sort(), async (app) => {

    await handler(app, argument)

  })

}

const _log = (text = '') => {

  if(process.env.NODE_ENV === 'test') return

  writePaddedLine(chalk.grey(text), '', '#FFFFFF', true, false)

}

const _getRepository = (name) => {

  if(name.match(/^https:\/\/github\.com\/.*\.git$/)) return name

  if(name.match(/^[\w\d-]*\/[\w\d-]*$/)) return `https://github.com/${name}.git`

  return `https://github.com/mahaplatform/${name}.git`

}

const _getRepositoryString = (name) => {

  const full = name.match(/^https:\/\/github\.com\/([\w\d-]*)\/([\w\d-]*)\.git$/)

  if(full) return `github:${full[1]}/${full[2]}`

  const namespaced = name.match(/^([\w\d-]*):([\w\d-]*)\/([\w\d-]*)$/)

  if(namespaced) return name

  return `github:mahaplatform/${name}`

}

const _cloneRepository = async (repository) => {

  const parts = repository.match(/\/([\w\d-]*)\.git/)

  if(!parts) throw new Error(`Unable to clone ${repository}`)

  const name = parts[1]

  const destination = path.join(appsPath, name)

  if(fs.existsSync(destination)) throw new Error(`${repository} already cloned`)

  await exec(`git clone ${repository}`, appsPath)

  return name

}

export const _cleanApp = async (app) => {

  _log(`Cleaning ${app}`)

  await exec(`rm -rf ${app}`, appsPath)

}

export const _installApp = async (app) => {

  if(fs.existsSync(path.join(appsPath, app))) {

    _log(`Updating ${app}`)

    return await exec('git pull', path.join(appsPath, app))

  }

  _log(`Installing ${app}`)

  const repository = _getRepository(app)

  await _cloneRepository(repository)

}

const _pullApp = async (name) => {

  _log(`Pulling app ${name}`)

  const appPath = path.join(appsPath, name)

  await exec('git pull', appPath)

}

const _commitApp = async (name, message) => {

  _log(`Commiting app ${name}`)

  const appPath = path.join(appsPath, name)

  await exec('git add .', appPath)

  await exec(`git commit -m '${message}'`, appPath)

  await exec('git push', appPath)

}

const _addApp = async (app) => {

  const repository = _getRepository(app)

  _log(`Adding ${repository}`)

  await _cloneRepository(repository)

  const config = _readConfig()

  config.apps[app] = _getRepositoryString(app)

  _writeConfig(config)

}

const _removeApp = async (app) => {

  _log(`Removing ${app}`)

  await exec(`rm -rf ${app}`, appsPath)

  const config = _readConfig()

  delete config.apps[app]

  _writeConfig(config)

}

export const addTask = async (flags, args) => {

  await _addApp(args.app)

}

export const removeTask = async (flags, args) => {

  await _removeApp(args.app)

}

export const installTask = async (flags, args) => {

  await executeForOneOrAll(_installApp, args.app)

}

export const cleanTask = async (flags, args) => {

  await executeForOneOrAll(_cleanApp, args.app)

}

export const commitTask = async (flags, args) => {

  await executeForOneOrAll(_commitApp, args.app, args.message)

}

export const pullTask = async (flags, args) => {

  await executeForOneOrAll(_pullApp, args.app)

}
