import sshPool from 'ssh-pool'
import path from 'path'
import _ from 'lodash'
import chalk from 'chalk'
import moment from 'moment'

const values = require(path.resolve('config', 'deploy.js'))

const config = {
  ...values.default,
  ...values.production
}

const timestamp = moment().format('YYYYMMDDHHmmss')

const sharedPath = path.join(config.deployTo, 'shared')

const currentPath = path.join(config.deployTo, 'current')

const releasePath = path.join(config.deployTo, 'releases', timestamp)

const servers = config.servers.reduce((servers, server) => {

  const fullname = `${server.user}@${server.host}`

  return {
    ...servers,
    all: [
      ...servers.all || [],
      ..._.includes(servers.all, fullname) ? [] : [fullname]
    ],
    ..._.castArray(server.role).reduce((roles, role) => ({
      ...roles,
      [role]: [
        ...servers[role] || [],
        fullname
      ]
    }), {})
  }

}, {})

const options = {
  stdout: process.stdout,
  stderr: process.stderr,
  log: console.log.bind(console),
  key: config.key
}

const getPool = (roles) => {

  if(!roles) return new sshPool.Connection({
    remote: 'root@localhost',
    ...options
  })

  const hosts = _.castArray(roles).reduce((hosts, role) => [
    ...hosts,
    ...servers[role].filter(host => !_.includes(hosts, host))
  ], [])

  return new sshPool.ConnectionPool(hosts, options)

}

const local = async (connection, cmd) => {

  await connection.runLocally(cmd)

}

const run = async (pool, cmd) => {

  await pool.run(cmd)

}

const sync = async (pool, paths) => {

  await pool.copyToRemote(paths[0], paths[1], {
    rsync: [
      '--delete'
    ]
  })

}

const getAction = (action) => {

  if(action === 'run') return run

  if(action === 'sync') return sync

  if(action === 'local') return local

}

const getDuration = (startTime) => {

  const diff = process.hrtime(startTime)

  const ms = (diff[0] * 1e3 + diff[1] * 1e-6) / 1000

  return `${ms.toFixed(3)}s`

}

const getFunction = (options, direction) => {

  if(!options[direction]) return () => {}

  const fn = getAction(options.action)

  const pool = getPool(options.roles)

  return async () => {

    const startTime = process.hrtime()

    const action = direction === 'up' ? 'DEPLOY' : 'ROLLBACK'

    process.stdout.write(chalk.green(`\n[${action}]`) + chalk.white(`: ${options.description}\n`))

    await fn(pool, options[direction])

    process.stdout.write(chalk.red(`Completed in ${getDuration(startTime)}\n`))

  }

}

const task = (options) => ({
  up: getFunction(options, 'up'),
  down: getFunction(options, 'down')
})

const makeReleasePath = task({
  description: 'Making release path',
  action: 'run',
  up: `rm -rf ${releasePath} && mkdir -p ${releasePath}`,
  down: `rm -rf ${releasePath}`,
  roles: ['appserver', 'worker', 'cron']
})

const copyPackage = task({
  description: 'Copying package files',
  action: 'sync',
  up: [
    path.join('package*json'),
    releasePath
  ],
  roles: ['appserver', 'worker', 'cron']
})

const copyNodeModules = task({
  description: 'Copying previous node_modules',
  action: 'run',
  up: `cp -rp ${currentPath}/node_modules ${releasePath}/node_modules`,
  roles: ['appserver', 'worker', 'cron']
})

const installNewModules = task({
  description: 'Installing new node_modules',
  action: 'run',
  up: `cd ${currentPath} && npm install`,
  roles: ['appserver', 'worker', 'cron']
})

const copyConfig = task({
  description: 'Syncing config',
  action: 'sync',
  up: [
    path.join('config', 'maha.production.js'),
    path.join(releasePath, 'maha.js')
  ],
  roles: ['appserver', 'worker', 'cron']
})


const compile = task({
  description: 'Compile source',
  action: 'local',
  up: 'NODE_ENV=production maha compile'
})

const syncCompiled = task({
  description: 'Syncing newly compiled src',
  action: 'sync',
  up: [
    path.join('dist'),
    releasePath
  ],
  roles: ['appserver', 'worker', 'cron']
})

const buildHelp = task({
  description: 'Build help files',
  action: 'local',
  up: 'NODE_ENV=production maha help:build'
})

const syncHelp = task({
  description: 'Syncing newly compiled help',
  action: 'sync',
  up: [
    path.join('help'),
    releasePath
  ],
  roles: ['appserver']
})

const linkImagecache = task({
  description: 'Linking imagecahe directory',
  action: 'run',
  up: `ln -s ${sharedPath}/imagecache ${releasePath}/dist/public/imagecache`,
  roles: 'appserver'
})

const linkLogs = task({
  description: 'Linking log directory',
  action: 'run',
  up: `ln -s ${sharedPath}/logs ${releasePath}/logs`,
  roles: ['appserver', 'worker', 'cron']
})

const symlinkRelease = task({
  description: 'Symlnk release',
  action: 'run',
  up: `rm -rf ${currentPath} && ln -s ${currentPath} ${releasePath}`,
  roles: ['appserver', 'worker', 'cron']
})

const bootstrap = task({
  description: 'Bootstrapping maha',
  action: 'run',
  up: `cd ${currentPath} && NODE_ENV=production maha bootstrap`,
  roles: 'controller'
})

const restartLoadBalancer = task({
  description: 'Restart load balancer',
  action: 'run',
  up: 'service haproxy restart',
  roles: 'loadbalancer'
})

const reloadAppServers = task({
  description: 'Reload appservers',
  action: 'run',
  up: `touch ${currentPath}/tmp/restart.txt`,
  roles: 'appserver'
})

const restartAppServers = task({
  description: 'Reload appservers',
  action: 'run',
  up: 'service nginx restart',
  roles: 'appserver'
})

const cacheAppservers = task({
  description: 'Cache servers',
  action: 'run',
  up: 'wget -O - http://127.0.0.1:80/ping && wget -O - http://127.0.0.1:81/ping',
  roles: 'appserver'
})

const restartWorkers = task({
  description: 'Restart workers',
  action: 'run',
  up: `cd ${currentPath} && NODE_ENV=production pm2 startOrRestart ./config/ecosystem.config.js`,
  roles: 'worker'
})

const restartCron = task({
  description: 'Restart cron',
  action: 'run',
  up: `cd ${currentPath} && NODE_ENV=production pm2 startOrRestart ./config/ecosystem.config.js`,
  roles: 'cron'
})

export const deploy = async (flags, args) => {

  const tasks = [
    makeReleasePath,
    copyPackage,
    copyNodeModules,
    installNewModules,
    copyConfig,
    compile,
    syncCompiled,
    buildHelp,
    syncHelp,
    linkImagecache,
    linkLogs,
    symlinkRelease,
    bootstrap,
    reloadAppServers,
    cacheAppservers,
    restartWorkers,
    restartCron
  ]

  let completed = []

  const startTime = process.hrtime()

  try {

    await Promise.mapSeries(tasks, async (task) => {

      await task.up()

      completed.unshift(task)

    })


  } catch(err) {

    process.stderr.write(err)

    await Promise.mapSeries(completed, async (task) => {

      await task.down()

    })

  }

  process.stdout.write(chalk.green(`\nCompleted in total ${getDuration(startTime)}\n`))

}
