import { writePaddedLine } from '../../utils/console'
import pluralize from 'pluralize'
import exec from '../../utils/exec'
import moment from 'moment'
import chalk from 'chalk'
import path from 'path'
import _ from 'lodash'
import ejs from 'ejs'
import fs from 'fs'

const modulesPath = path.resolve('node_modules')

const appsPath = path.resolve('app')

const generatorPath = path.resolve(__dirname, '..', '..', '..', 'src','generators')

const generators = fs.readdirSync(generatorPath)

export const Generate = async (flags, args) => {

  const template = args.template

  const app = args.app

  const name = typeof args.name === 'string' ? args.name : args.app

  const generatorPath = generators.find(generator => generator == template)

  if(!generatorPath) throw new Error('invalid template')

  const templatesPath = path.join(generatorPath, template)

  const appConfigFilePath = path.join(_getAppRoot(app), app, 'src', 'app.js')

  const appConfig = fs.existsSync(appConfigFilePath) ? require(appConfigFilePath).default.config : null

  const data = getData(template, app, appConfig, name)

  const generator = require(generatorFilePath).default

  generator.files.map(file => generate(file, templatesPath, data))

  if(generator.after) await runHooks(generator.after, data)

}

export const Destroy = (template, app, name) => {

  const generatorFilePath = path.join(modulesPath, 'maha', 'dist', 'generators', template, 'index.js')

  if(!fs.existsSync(generatorFilePath)) return

  const appConfigFilePath = path.join(modulesPath, app, 'dist', 'app.js')

  const appConfig = fs.existsSync(appConfigFilePath) ? require(appConfigFilePath).default.config : null

  const data = getData(template, appConfig, name)

  const generator = require(generatorFilePath).default

  generator.files.map(file => destroy(file, data))

}

const generate = (options, templatesPath, data) => {

  const appPath = ejs.render(`${data.app}/${options.filepath}`, data)

  const fullPath = appPath.split('/').slice(0,-1).reduce((fullPath, segment) => {

    const joinedPath = path.join(fullPath, segment)

    const directoryPath = path.join(data.root, joinedPath)

    if(!fs.existsSync(directoryPath)) {

      writePaddedLine(`${chalk.green('mkdir')}     ${joinedPath}`)

      fs.mkdirSync(directoryPath)

    }

    return joinedPath

  }, '')

  const rendered = getRendered(options, templatesPath, data)

  const filename = ejs.render(options.filepath.split('/').slice(-1)[0], data)

  const filePath = path.join(fullPath, filename)

  const exists = fs.readdirSync(path.join(data.root, fullPath)).reduce((exists, file) => {

    return exists || file.replace(data.timestamp, '').search(filename.replace(data.timestamp, '')) >= 0

  }, false)

  if(options.action === 'create') {

    if(exists) return writePaddedLine(`${chalk.hex('#2185D0')('identical')} ${filePath}`)

    writePaddedLine(`${chalk.green('write')}     ${filePath}`)

    fs.writeFileSync(path.join(data.root, filePath), rendered)

  } else if(options.action === 'prepend') {

    if(!exists) return writePaddedLine(`${chalk.hex('#DB2828')('notfound')}    ${filePath}`)

    const contents = fs.readFileSync(path.resolve(filePath)).toString()

    const prepended = options.before ? contents.replace(options.before, `${rendered}${options.before}`) : `${rendered}\n${contents}`

    fs.writeFileSync(path.join(data.root, filePath), prepended)

    writePaddedLine(`${chalk.green('alter')}    ${filePath}`)

  } else if(options.action === 'append') {

    if(!exists) return writePaddedLine(`${chalk.hex('#DB2828')('notfound')}    ${filePath}\n`)

    const contents = fs.readFileSync(path.resolve(filePath)).toString()

    const appended = options.after ? contents.replace(options.after, `${options.after}${rendered}`) : `${contents}\n${rendered}`

    fs.writeFileSync(path.join(data.root, filePath), appended)

    writePaddedLine(`${chalk.green('alter')}    ${filePath}`)

  }

}

const runHooks = async (hooks, data) => {

  return await Promise.mapSeries(hooks, async (hook) => {

    writePaddedLine(`${chalk.green('exec')}      ${hook.description}`)

    await exec(hook.command, `${data.root}/${data.path}`)

  })

}

const destroy = (options, data) => {

  const fullPath = options.filepath.split('/').slice(0,-1).reduce((fullPath, segment) => {

    const interpolated = ejs.render(segment, data)

    return path.join(fullPath, interpolated)

  }, '')

  const filename = ejs.render(options.filepath.split('/').slice(-1)[0], data)

  const filePath = path.join(fullPath, filename)

  if(!fs.existsSync(path.resolve(filePath))) return

  writePaddedLine(`${chalk.green('delete')}     ${filePath}`)

  fs.unlinkSync(path.resolve(filePath))

}

const getRendered = (options, templatesPath, data) => {

  if(options.template) {

    const templateFilePath = path.join(templatesPath, options.template)

    const template = fs.readFileSync(templateFilePath, 'utf8')

    return ejs.render(template, data)

  } else if(options.content) {

    return ejs.render(options.content, data)

  } else {

    return ''

  }

}

const getData = (template, app, appConfig, initname) => {

  const parts = initname.split('/').map(_.snakeCase)

  const name = parts[parts.length - 1]

  const app_name = (_.includes(['platform','app'], template)) ? _.kebabCase(name) : appConfig.title

  return {
    _,
    app,
    name,
    app_name,
    root: template == 'app' ? appsPath : _getAppRoot(app),
    path: parts.join('/'),
    pluralize,
    timestamp: moment().format('YYYYMMDDHHmmss')
  }

}

const _getAppRoot = (app) => {

  return fs.existsSync(path.join(modulesPath, app)) ? modulesPath : appsPath

}
