import { writePaddedLine } from '../../utils/console'
import { build } from './build'
import exec from '../../utils/exec'
import rimraf from 'rimraf'
import chalk from 'chalk'
import path from 'path'
import _ from 'lodash'
import fs from 'fs'

const platformPath = path.resolve('.')

const packagesPath = path.join(platformPath, 'packages')

const modulesPath = path.join(platformPath, 'node_modules')

const cachedModules = {}

const _getModules = (modulesPath) => {

  if(cachedModules[modulesPath]) return cachedModules[modulesPath]

  cachedModules[modulesPath] = fs.readdirSync(modulesPath).reduce((modules, moduleName) => {

    const modulePath = path.join(modulesPath, moduleName)

    if(!fs.statSync(modulePath).isDirectory()) return modules

    return {
      ...modules,
      [moduleName]: modulePath
    }

  }, {})

  return cachedModules[modulesPath]

}

const _validatePackage = (name) => {

  const packages = Object.keys(_getModules('packages'))

  if(!_.includes(packages, name)) throw new Error(`Package ${name} is not installed`)

}

const executeForOneOrAll = async (handler, name, arg) => {

  const argument = (typeof arg === 'string') ? arg : null

  if(typeof name === 'string' && name !== 'all') {

    _validatePackage(name)

    return await handler(name, argument)

  }

  const packages = Object.keys(_getModules(packagesPath)).sort()

  await Promise.map(packages, (name) => handler(name, argument))

}

const _log = (text = '') => {

  if(process.env.NODE_ENV === 'test') return

  writePaddedLine(chalk.grey(text), '', '#FFFFFF', true, false)

}

const _getRepository = (name) => {

  if(name.match(/^https:\/\/github\.com\/.*\.git$/)) return name

  if(name.match(/^[\w\d-]*\/[\w\d-]*$/)) return `https://github.com/${name}.git`

  return `https://github.com/mahaplatform/${name}.git`

}

const _getDependencies = (packagePath) => {

  const configPath = path.join(packagePath, 'package.json')

  if(!fs.existsSync(configPath)) return []

  const config = JSON.parse(fs.readFileSync(configPath).toString())

  return [
    ...config.devDependencies ? Object.keys(config.devDependencies) : [],
    ...config.dependencies ? Object.keys(config.dependencies) : []
  ]

}

const _clonePackage = async (repo) => {

  const repository = _getRepository(repo)

  _log(`Cloning ${repository}`)

  const parts = repository.match(/\/([\w\d-]*)\.git/)

  if(!parts) throw new Error(`Unable to clone ${repository}`)

  const name = parts[1]

  const destination = path.join(packagesPath, name)

  if(fs.existsSync(destination)) throw new Error(`${repository} already cloned`)

  await exec(`git clone ${repository}`, packagesPath)

  return name

}

const _installPackage = async (name, dependency) => {

  const description = dependency ? dependency : 'all modules'

  _log(`Installing ${description} for package ${name}`)

  const packagePath = path.join(packagesPath, name)

  const command = dependency ? `npm install ${dependency}` : 'npm install'

  await exec(command, packagePath)

}

const _cleanPackage = async (name) => {

  _log(`Cleaning package ${name}`)

  const packagePath = path.join(packagesPath, name)

  await exec('rm -rf dist', packagePath)

}

const _pullPackage = async (name) => {

  _log(`Pulling package ${name}`)

  const packagePath = path.join(packagesPath, name)

  await exec('git pull', packagePath)

}


const _buildPackage = async (name) => {

  _log(`Building package ${name}`)

  await build({}, { root: 'packages', app: name })

}


const _linkPackage = async (name) => {

  _log(`Linking package ${name}`)

  const packagePath = path.join(packagesPath, name)

  const modulePath = path.join(modulesPath, name)

  if(modulePath) {

    rimraf.sync(modulePath)

    fs.symlinkSync(packagePath, modulePath)

  }

  const packages = _getModules(packagesPath)

  _getDependencies(packagePath).map(dependency => {

    if(!packages[dependency]) return

    const packageModulePath = path.join(packagePath, 'node_modules', dependency)

    if(fs.statSync(packageModulePath).isDirectory()) rimraf.sync(packageModulePath)

    fs.symlinkSync(packages[dependency], packageModulePath)

  })

  fs.readdirSync(modulesPath).map(moduleName => {

    if(!packages[moduleName]) return

    const modulePath = path.join(modulesPath, moduleName)

    if(fs.statSync(modulePath).isDirectory()) rimraf.sync(modulePath)

    fs.symlinkSync(packages[moduleName], modulePath)

  })

}

const _commitPackage = async (name, message) => {

  _log(`Commiting package ${name}`)

  const packagePath = path.join(packagesPath, name)

  await exec('git add .', packagePath)

  await exec(`git commit -m '${message}'`, packagePath)

  await exec('git push', packagePath)

}

export const importTask = async (flags, args) => {

  const name = await _clonePackage(args.repository)

  await _installPackage(name)

  await _buildPackage(name)

  await _linkPackage(name)

}

export const installTask = async (name, dependency) => {

  await executeForOneOrAll(_installPackage, name, dependency)

}

export const buildTask = async (name) => {

  await executeForOneOrAll(_buildPackage, name)

}

export const linkTask = async (name) => {

  await executeForOneOrAll(_linkPackage, name)

}

export const commitTask = async (name, message) => {

  await executeForOneOrAll(_commitPackage, name, message)

}

export const cleanTask = async (name, message) => {

  await executeForOneOrAll(_cleanPackage, name, message)

}

export const pullTask = async (name, message) => {

  await executeForOneOrAll(_pullPackage, name, message)

}
